@@ -7,17 +7,17 @@
 # modify it under the terms of the GNU Lesser General Public
 # License as published by the Free Software Foundation; either
 # version 2.1 of the License, or (at your option) any later version.
-# 
+#
 # This library is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 # Lesser General Public License for more details.
-# 
+#
 # You should have received a copy of the GNU Lesser General Public
 # License along with this library; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 # USA
-# 
+#
 # Author(s): Gustavo J. A. M. Carneiro <gjc@inescporto.pt>
 #
 
@@ -87,7 +87,7 @@ Syntax for yield in tasklets
        - L{WaitCondition}, meaning to wait for that specific condition
 
        - L{Tasklet}, with the same meaning as L{WaitForTasklet}C{(tasklet)}
-       
+
        - generator, with the same meaning as L{WaitForTasklet}C{(Tasklet(gen))}
 
       In this case, the tasklet is suspended until either one of the
@@ -183,8 +183,6 @@ import warnings
 
 import gobject
 
-from kiwi.argcheck import argcheck
-
 if gobject.pygtk_version <= (2, 8):
     raise RuntimeError("PyGTK 2.8 or later is required for kiwi.tasklet")
 
@@ -198,7 +196,7 @@ class task(object):
         self._func = func
         self.__name__ = func.__name__
         self.__doc__ = func.__doc__
-        
+
     def __call__(self, *args, **kwargs):
         return Tasklet(self._func(*args, **kwargs))
 
@@ -235,7 +233,7 @@ class WaitCondition(object):
     WaitConditions are used in a yield statement inside tasklets body
     for specifying what event(s) it should wait for in order to
     receive control once more.'''
-    
+
     def __init__(self):
         '''Abstract base class, do not call directly'''
         self.triggered = False
@@ -267,6 +265,57 @@ class WaitCondition(object):
         raise NotImplementedError
 
 
+class WaitForCall(WaitCondition):
+    '''An object that waits until it is called.
+
+      This example demonstrates how a tasklet waits for a callback::
+        import gobject
+        from kiwi import tasklet
+
+        mainloop = gobject.MainLoop()
+
+        def my_task():
+            callback = tasklet.WaitForCall()
+            gobject.timeout_add(1000, callback)
+            yield callback
+            mainloop.quit()
+
+        tasklet.run(my_task())
+        mainloop.run()
+
+      @ivar return_value: value to return when called
+
+    '''
+    def __init__(self, return_value=None):
+        '''
+        Creates a wait condition that is actually a callable object, and waits for a call to be made on it.
+        @param return_value: value to return when called; can also be
+        modified dynamically from the tasklet as the C{return_value}
+        instance variable.
+        '''
+        WaitCondition.__init__(self)
+        self.return_value = return_value
+        self.args = None
+        self.kwargs = None
+        self._callback = None
+
+    def arm(self, tasklet):
+        '''Overrides WaitCondition.arm'''
+        self._callback = tasklet.wait_condition_fired
+
+    def disarm(self):
+        '''Overrides WaitCondition.disarm'''
+        self._callback = None
+
+    def __call__(self, *args, **kwargs):
+        self.triggered = True
+        self.args = args
+        self.kwargs = kwargs
+        retval = self._callback(self)
+        self.triggered = False
+        return self.return_value
+
+
 class WaitForIO(WaitCondition):
     '''An object that waits for IO conditions on sockets or file
     descriptors.
@@ -283,7 +332,7 @@ class WaitForIO(WaitCondition):
         @type condition: a set of C{gobject.IO_*} flags ORed together
         @param priority: mainloop source priority
         '''
-        
+
         WaitCondition.__init__(self)
         self.filedes = filedes
         self._condition = condition # listen condition
@@ -339,13 +388,13 @@ class WaitForTimeout(WaitCondition):
         @param timeout: ammount of time to wait, in miliseconds
         @param priority: mainloop priority for the timeout event
         '''
-        
+
         WaitCondition.__init__(self)
         self.timeout = timeout
         self._id = None
         self._tasklet = None
         self._priority = priority
-        
+
     def arm(self, tasklet):
         '''See L{WaitCondition.arm}'''
         if self._id is None:
@@ -412,14 +461,14 @@ class WaitForTasklet(WaitCondition):
     '''An object that waits for a tasklet to complete'''
     def __init__(self, tasklet):
         '''An object that waits for another tasklet to complete'''
-        
+
         WaitCondition.__init__(self)
         self._tasklet = tasklet
         self._id = None
         self._idle_id = None
         self._callback = None
         self.retval = None
-        
+
     def arm(self, tasklet):
         '''See L{WaitCondition.arm}'''
         self._callback = tasklet.wait_condition_fired
@@ -456,7 +505,6 @@ class WaitForTasklet(WaitCondition):
 class WaitForSignal(WaitCondition):
     '''An object that waits for a signal emission'''
 
-    @argcheck(gobject.GObject, str)
     def __init__(self, obj, signal):
         '''Waits for a signal to be emitted on a specific GObject instance.
 
@@ -466,6 +514,11 @@ class WaitForSignal(WaitCondition):
         @type signal: str
         '''
         WaitCondition.__init__(self)
+        if not isinstance(obj, gobject.GObject):
+            raise TypeError("obj must be a GObject instance")
+        if not gobject.signal_lookup(signal, obj):
+            raise ValueError("gobject %r does not have a signal called %r" %
+                             (obj, signal))
         self.object = obj
         self.signal = signal
         self._callback = None
@@ -481,7 +534,7 @@ class WaitForSignal(WaitCondition):
             if gobject.signal_lookup("destroy", self.object):
                 self._destroy_id = self.object.connect("destroy",
                                                        self._object_destroyed)
-                
+
     def _object_destroyed(self, obj):
         self.object = None
         self._id = None
@@ -548,7 +601,7 @@ class WaitForProcess(WaitCondition):
 
 class Message(object):
     '''A message that can be received by or sent to a tasklet.'''
-    
+
     _slots_ = 'name', 'dest', 'value', 'sender'
 
     ACCEPT, DEFER, DISCARD = range(3)
@@ -571,7 +624,7 @@ class Message(object):
         self.value = value
         self.sender = sender
         self.dest = dest
-        
+
 #     def get_name(self):
 #         """Return the message name"""
 #         return self.name
@@ -590,7 +643,7 @@ def _normalize_list_argument(arg, name):
     """returns a list of strings from an argument that can be either
     list of strings, None (returns []), or a single string returns
     ([arg])"""
-    
+
     if arg is None:
         return []
     elif isinstance(arg, basestring):
@@ -632,7 +685,7 @@ class WaitForMessages(WaitCondition):
             self.actions[name] = Message.DEFER
         for name in discard:
             self.actions[name] = Message.DISCARD
-        
+
     def arm(self, tasklet):
         '''Overrides WaitCondition.arm'''
         self._tasklet = tasklet
@@ -646,10 +699,20 @@ class WaitForMessages(WaitCondition):
 
 
 class Tasklet(object):
-    '''An object that launches and manages a tasklet.'''
+    '''An object that launches and manages a tasklet.
+
+    @ivar state: current execution state of the tasklet, one of the STATE_* contants.
+
+    @ivar return_value: the value returned by the task function, or None.
+
+    @cvar STATE_RUNNING: the tasklet function is currently executing code
+    @cvar STATE_SUSPENDED: the tasklet function is currently waiting for an event
+    @cvar STATE_MSGSEND: the tasklet function is currently sending a message
+    @cvar STATE_ZOMBIE: the tasklet function has ended
+    '''
+
+    STATE_RUNNING, STATE_SUSPENDED, STATE_MSGSEND, STATE_ZOMBIE = range(4)
 
-    STATE_RUNNING, STATE_SUSPENDED, STATE_MSGSEND = range(3)
-    
     def __init__(self, gen=None):
         '''
         Launch a generator tasklet.
@@ -666,6 +729,7 @@ class Tasklet(object):
         self._message_queue = []
         self._message_actions = {}
         self.state = Tasklet.STATE_SUSPENDED
+        self.return_value = None
         if gen is None:
             self.gen = self.run()
         else:
@@ -681,7 +745,7 @@ class Tasklet(object):
         return self._message_actions
 
     message_actions = property(get_message_actions)
-        
+
     def run(self):
         """
         Method that executes the task.
@@ -696,18 +760,20 @@ class Tasklet(object):
         assert _event is None
         had_event = (self._event is not None)
         _event = self._event
+        self.state = Tasklet.STATE_RUNNING
         try:
-            self.state = Tasklet.STATE_RUNNING
             gen_value = self.gen.next()
-            self.state = Tasklet.STATE_SUSPENDED
-            assert gen_value is not None
         except StopIteration, ex:
+            self.state = Tasklet.STATE_ZOMBIE
             if ex.args:
                 retval, = ex.args
             else:
                 retval = None
             self._join(retval)
             return None
+        else:
+            self.state = Tasklet.STATE_SUSPENDED
+            assert gen_value is not None            
         if __debug__:
             if had_event and _event is not None:
                 warnings.warn("Tasklet %s forgot to read an event!" % self)
@@ -726,6 +792,7 @@ class Tasklet(object):
             if isinstance(gen_value, Message):
                 msg = gen_value
                 self.state = Tasklet.STATE_MSGSEND
+                msg.sender = self
                 msg.dest.send_message(msg)
                 continue # loop because we posted a message
             elif isinstance(gen_value, tuple):
@@ -734,7 +801,7 @@ class Tasklet(object):
                 self.wait_list = gen_value
             else:
                 self.wait_list = [gen_value]
-            
+
             for i, val in enumerate(self.wait_list):
                 if isinstance(val, WaitCondition):
                     continue
@@ -788,16 +855,16 @@ class Tasklet(object):
         else:
             return None
         return self._message_queue.pop(idx)
-    
+
 
     def _update_wait_conditions(self, old_wait_list):
         '''disarm wait conditions removed and arm new wait conditions'''
-        
+
         ## disarm conditions removed from the wait list
         for cond in old_wait_list:
             if cond not in self.wait_list:
                 cond.disarm()
-        
+
         ## arm the conditions added to the wait list
         for cond in self.wait_list:
             if cond not in old_wait_list:
@@ -844,8 +911,11 @@ class Tasklet(object):
     def _join(self, retval):
         for cond in self.wait_list:
             cond.disarm()
+
         self.gen = None
+        self.return_value = retval
         self.wait_list = []
+
         callbacks = self._join_callbacks.values()
         self._join_callbacks.clear()
         for callback in callbacks:
